;; Small hacky script to build the planet distribution.  I need to
;; filter out all my test code (anything file whose name begins with
;; "test-"

(require (lib "util.ss" "planet"))
(require (lib "file.ss"))


(define tmpdir (build-path "suffixtree"))


;; FIXME: use find-files instead!  I should have read (lib "file.ss")
;; earlier... but I'm lazy and will clean this up later.
(define (build-planet-distribution)
  (make-directory* tmpdir)
  (make-directory* (build-path tmpdir "private"))
  (copy-file* (build-path "doc.txt") tmpdir)
  (shallow-copy-source-files/excluding-tests (build-path "private") (build-path tmpdir "private"))
  (shallow-copy-source-files/excluding-tests (current-directory) tmpdir)
  (make-planet-archive tmpdir)
  (delete-directory/files tmpdir))


(define (make-directory* path)
  (unless (directory-exists? path)
    (make-directory path)))


(define (copy-file* file dest-path)
  (unless (file-exists? (build-path dest-path file))
    (copy-file file (build-path dest-path file))))


;; Returns #t if path is a source file that should be included in the
;; PLaneT distribution.
(define (ok-source-file-path? path)
  (let-values (((base name must-be-dir?) (split-path path)))
    (and (file-exists? path)
         (path-contains-module? path)
         (not (bytes-suffix? #"~" (path->bytes name)))
         (not (bytes-prefix? #"test-" (path->bytes name))))))


(define (path-contains-module? path)
  (call-with-input-file path
    (lambda (ip)
      (let ((sexp (read ip)))
        (and (list? sexp)
             (not (null? sexp))
             (eq? 'module (car sexp)))))))
  

(define (bytes-prefix? possible-prefix s)
  (and (>= (bytes-length s) (bytes-length possible-prefix))
       (bytes=? possible-prefix (subbytes s 0 (bytes-length possible-prefix)))))

(define (bytes-suffix? possible-suffix s)
  (and (>= (bytes-length s) (bytes-length possible-suffix))
       (bytes=? possible-suffix (subbytes s (- (bytes-length s) (bytes-length possible-suffix))))))


(define (shallow-copy-source-files/excluding-tests source-path dest-path)
  (let ((files (directory-list source-path)))
    (for-each (lambda (f)
                (when (and (ok-source-file-path? (build-path source-path f))
                           (not (file-exists? (build-path dest-path f))))
                  (copy-file (build-path source-path f) (build-path dest-path f))))
              files)))



(build-planet-distribution)
